from datetime import UTC, datetime
from json import JSONDecodeError
from typing import Any

import typer
from tenacity import retry, retry_if_result, stop_never, wait_fixed

from jailrun.ansible import run_playbook
from jailrun.config import save_state
from jailrun.misc import lock
from jailrun.network import SSHKwargs, get_ssh_kw, jail_ssh_exec, ssh_exec, wait_for_ssh
from jailrun.qemu import vm_is_running
from jailrun.schemas import PeerJail, PeerState, State
from jailrun.serializers import dumps, loads
from jailrun.settings import Settings
from jailrun.ui import con, err, info, ok, warn


def _relay_request(
    ssh_kw: SSHKwargs, *, method: str, url: str, body: dict[str, Any] | None = None
) -> dict[str, Any] | None:
    cmd_parts = ["curl", "-s"]
    if method != "GET":
        cmd_parts += ["-X", method]
    if body is not None:
        cmd_parts += ["-H", '"Content-Type: application/json"', "-d", f"'{dumps(body)}'"]

    cmd_parts.append(f'"{url}"')

    raw = ssh_exec(cmd=" ".join(cmd_parts), ssh_kw=ssh_kw)
    if not raw:
        return None
    try:
        result: dict[str, Any] = loads(raw)
        return result
    except JSONDecodeError:
        return None


def _collect_jail_roster(ssh_kw: SSHKwargs, state: State) -> list[PeerJail]:
    roster: list[PeerJail] = []
    for name, jail in state.jails.items():
        if not jail.ip:
            continue

        raw = jail_ssh_exec(
            cmd="yggdrasilctl -json getSelf",
            jail_ip=jail.ip,
            ssh_kw=ssh_kw,
        )
        if not raw:
            continue

        try:
            data = loads(raw)
            addr = data.get("address", "")
            if addr:
                roster.append(PeerJail(name=name, ygg_address=addr))
        except (JSONDecodeError, KeyError):
            continue

    return roster


def pair_create(state: State, settings: Settings) -> None:
    with lock(settings.state_file):
        _pair_create(state=state, settings=settings)


def _pair_create(state: State, settings: Settings) -> None:
    alive, _ = vm_is_running(settings.pid_file)
    if not alive:
        err("VM is not running. Start it first with: jrun start")
        raise typer.Exit(1)

    ssh_kw = get_ssh_kw(settings, state)

    with con().status("[dim]Connecting to VM…[/dim]", spinner="dots"):
        wait_for_ssh(ssh_kw, silent=True)

    with con().status("[dim]Collecting jail addresses…[/dim]", spinner="dots"):
        roster = _collect_jail_roster(ssh_kw, state)

    if not roster:
        err("No mesh-enabled jails found. Deploy jails first with: jrun up")
        raise typer.Exit(1)

    data = _relay_request(
        ssh_kw,
        method="POST",
        url=f"https://{settings.relay_peer_uri}/pair",
        body={"jails": [j.model_dump() for j in roster]},
    )
    if not data or "code" not in data:
        err("Failed to create pairing on relay.")
        raise typer.Exit(1)

    code = data["code"]

    c = con()
    c.print()
    c.print(f"[bold cyan]Code:[/bold cyan]  {code}")
    c.print(f"[dim]Tell your peer:[/dim]  jrun pair {code}")
    c.print()

    @retry(stop=stop_never, wait=wait_fixed(3), retry=retry_if_result(lambda r: r is None))
    def _poll() -> dict[str, Any] | None:
        resp = _relay_request(
            ssh_kw,
            method="GET",
            url=f"https://{settings.relay_peer_uri}/pair/{code}",
        )
        if not resp:
            return None
        if resp.get("joined"):
            return resp
        return None

    try:
        with c.status("[dim]Waiting for peer…[/dim]", spinner="dots"):
            result = _poll()
    except KeyboardInterrupt:
        c.print()
        warn("Cancelled.")
        raise typer.Exit(0) from None

    if result is None:
        err("Pairing failed.")
        raise typer.Exit(1)

    peer = PeerState(
        alias=code,
        direction="init",
        paired_at=datetime.now(UTC).isoformat(),
        jails=[PeerJail(**j) for j in result["jails"]],
    )
    state.peers.append(peer)
    save_state(state=state, state_file=settings.state_file)

    ok(f"Paired with {code} ({len(peer.jails)} jails)")
    _apply_peers(state=state, settings=settings)


def _apply_peers(state: State, settings: Settings) -> None:
    peers_data = [p.model_dump() for p in state.peers]

    info("Updating DNS records…")
    run_playbook(
        "jail-dns-peers.yml",
        extra_vars={"peers": peers_data},
        settings=settings,
        state=state,
    )

    info("Updating pf firewall rules…")
    run_playbook(
        "jail-pf-peers.yml",
        extra_vars={"peers": peers_data},
        settings=settings,
        state=state,
    )


def pair_join(code: str, state: State, settings: Settings) -> None:
    with lock(settings.state_file):
        _pair_join(code=code, state=state, settings=settings)


def _pair_join(code: str, state: State, settings: Settings) -> None:
    alive, _ = vm_is_running(settings.pid_file)
    if not alive:
        err("VM is not running. Start it first with: jrun start")
        raise typer.Exit(1)

    ssh_kw = get_ssh_kw(settings, state)

    with con().status("[dim]Connecting to VM…[/dim]", spinner="dots"):
        wait_for_ssh(ssh_kw, silent=True)

    with con().status("[dim]Collecting jail addresses…[/dim]", spinner="dots"):
        roster = _collect_jail_roster(ssh_kw, state)

    if not roster:
        err("No mesh-enabled jails found. Deploy jails first with: jrun up")
        raise typer.Exit(1)

    data = _relay_request(
        ssh_kw,
        method="POST",
        url=f"https://{settings.relay_peer_uri}/pair/{code}",
        body={"jails": [j.model_dump() for j in roster]},
    )
    if data is None:
        err("Invalid or expired code.")
        raise typer.Exit(1)

    peer = PeerState(
        alias=code,
        direction="joined",
        paired_at=datetime.now(UTC).isoformat(),
        jails=[PeerJail(**j) for j in data["jails"]],
    )
    state.peers.append(peer)
    save_state(state=state, state_file=settings.state_file)

    ok(f"Paired with {code} ({len(peer.jails)} jails)")
    _apply_peers(state=state, settings=settings)


def pair_list(state: State) -> None:
    from rich.padding import Padding
    from rich.table import Table
    from rich.text import Text

    c = con()
    c.print()

    if not state.peers:
        c.print("  [dim]no paired instances[/dim]")
        c.print()
        return

    tbl = Table(
        show_header=True,
        header_style="dim",
        box=None,
        show_edge=False,
        padding=(0, 3, 0, 0),
        expand=False,
        pad_edge=False,
    )
    tbl.add_column("name", no_wrap=True)
    tbl.add_column("peer", style="dim white", no_wrap=True)
    tbl.add_column("ip", style="dim white", no_wrap=True)
    tbl.add_column("paired", style="dim white", no_wrap=True)

    for p in state.peers:
        for j in p.jails:
            tbl.add_row(
                Text(j.name, style="bold"),
                p.alias,
                j.ygg_address,
                datetime.fromisoformat(p.paired_at).strftime("%b %d %Y, %H:%M UTC"),
            )

    c.print(Padding(tbl, pad=(0, 0, 0, 2)))
    c.print()


def pair_remove(alias: str, state: State, settings: Settings) -> None:
    with lock(settings.state_file):
        _pair_remove(alias=alias, state=state, settings=settings)


def _pair_remove(alias: str, state: State, settings: Settings) -> None:
    match = [p for p in state.peers if p.alias == alias]
    if not match:
        err(f"No peer found with name '{alias}'")
        raise typer.Exit(1)

    state.peers = [p for p in state.peers if p.alias != alias]
    save_state(state=state, state_file=settings.state_file)

    ok(f"Removed {alias}")
    _apply_peers(state=state, settings=settings)



from jailrun.cmd.down import down as down
from jailrun.cmd.pair import pair_create as pair_create
from jailrun.cmd.pair import pair_join as pair_join
from jailrun.cmd.pair import pair_list as pair_list
from jailrun.cmd.pair import pair_remove as pair_remove
from jailrun.cmd.purge import purge as purge
from jailrun.cmd.snapshot import snapshot_create as snapshot_create
from jailrun.cmd.snapshot import snapshot_delete as snapshot_delete
from jailrun.cmd.snapshot import snapshot_list as snapshot_list
from jailrun.cmd.snapshot import snapshot_rollback as snapshot_rollback
from jailrun.cmd.ssh import run_cmd as run_cmd
from jailrun.cmd.ssh import ssh as ssh
from jailrun.cmd.start import start as start
from jailrun.cmd.status import status as status
from jailrun.cmd.stop import stop as stop
from jailrun.cmd.up import up as up



import time

from rich.live import Live
from rich.padding import Padding
from rich.table import Table
from rich.text import Text
from rich.tree import Tree

from jailrun.network import SSHKwargs, jail_ssh_exec
from jailrun.ui import con

from .monit import parse_monit_status
from .render import build_jail_info, build_jail_table, find_jail, jail_header
from .sparkline import SampleHistory
from .types import JailRow, MonitJailStatus, StatusInfo


def refresh_jail_monit(j: JailRow, *, ssh_kw: SSHKwargs) -> None:
    if j["state"].lower() != "up" or not j["ips"]:
        return

    jail_ip = j["ips"][0]
    try:
        raw = jail_ssh_exec(
            cmd="monit status 2>/dev/null",
            jail_ip=jail_ip,
            timeout=10,
            ssh_kw=ssh_kw,
        )
    except Exception:
        return

    if not raw:
        return

    parsed = parse_monit_status(raw)
    for _, monit_data in parsed.items():
        j["monit"] = MonitJailStatus(system_ok=None, services=[])

        if monit_data["system_ok"] is not None:
            j["monit"]["system_ok"] = monit_data["system_ok"]

        j["monit"]["services"].extend(monit_data["services"])


def _build_live_layout(
    j: JailRow,
    *,
    history: SampleHistory,
    interval: int,
    tree: bool = False,
) -> Table:
    layout = Table.grid(padding=0)
    layout.add_column()

    layout.add_row("")
    layout.add_row(jail_header(j))
    layout.add_row("")
    layout.add_row(Padding(build_jail_info(j), pad=(0, 0, 0, 2)))
    layout.add_row("")

    if tree:
        root = Tree(Text("services", style="bold"))
        monit = j.get("monit")
        if monit and monit["services"]:
            for svc in monit["services"]:
                svc_style = "green" if svc["status"] == "ok" else "red"
                label = Text.assemble(
                    (svc["name"], "bold white"),
                    ("  ", ""),
                    (svc["status"], svc_style),
                )
                node = root.add(label)
                if svc["cpu"]:
                    line = Text.assemble(("cpu     ", "dim cyan"), (svc["cpu"].ljust(10), "white"))
                    line.append_text(history.cpu_spark(j["name"], svc["name"]))
                    node.add(line)

                if svc["mem"]:
                    line = Text.assemble(("mem     ", "dim cyan"), (svc["mem"].ljust(10), "white"))
                    line.append_text(history.mem_spark(j["name"], svc["name"]))
                    node.add(line)

                if svc["uptime"]:
                    node.add(Text.assemble(("uptime  ", "dim cyan"), (svc["uptime"], "dim white")))

        layout.add_row(Padding(root, pad=(0, 0, 0, 2)))
    else:
        layout.add_row(Padding(build_jail_table(j, history=history), pad=(0, 0, 0, 2)))

    layout.add_row("")
    layout.add_row(Text(f"  refreshing every {interval}s · Ctrl-C to exit", style="dim white"))
    layout.add_row("")

    return layout


def render_jail_live(
    data: StatusInfo,
    jail_name: str,
    *,
    ssh_kw: SSHKwargs,
    tree: bool = False,
    interval: int = 5,
) -> None:
    j = find_jail(data, jail_name)

    history = SampleHistory()
    history.ingest_jail(j)

    c = con()

    try:
        with Live(
            _build_live_layout(j, history=history, interval=interval, tree=tree),
            console=c,
            refresh_per_second=1,
            screen=False,
        ) as live:
            while True:
                time.sleep(interval)
                refresh_jail_monit(j, ssh_kw=ssh_kw)
                history.ingest_jail(j)
                live.update(_build_live_layout(j, history=history, interval=interval, tree=tree))
    except KeyboardInterrupt:
        c.print()
        c.print(Text("  Stopped.", style="dim white"))
        c.print()



import re
from dataclasses import dataclass, field
from typing import Literal

from .types import MonitJailStatus, MonitService

BlockType = Literal["system", "process", "program"]

HEADER_RE = re.compile(r"^(System|Process|Program)\s+'(.+)'\s*$")

COMPOUND_KEYS = (
    "cpu total",
    "memory usage",
    "monitoring status",
    "monitoring mode",
)


@dataclass
class MonitBlock:
    type: BlockType
    name: str
    fields: dict[str, str] = field(default_factory=dict)


def _parse_field_line(line: str) -> tuple[str, str] | None:
    lower = line.lower()
    for ck in COMPOUND_KEYS:
        if lower.startswith(ck):
            val = line[len(ck) :].strip()
            return (ck, val) if val else None

    parts = line.split(None, 1)

    return (parts[0].lower(), parts[1]) if len(parts) == 2 else None


def _tokenize(raw: str) -> list[MonitBlock]:
    blocks: list[MonitBlock] = []
    current: MonitBlock | None = None

    for line in raw.splitlines():
        stripped = line.strip()
        if not stripped:
            continue

        m = HEADER_RE.match(stripped)
        if m:
            current = MonitBlock(
                type=m.group(1).lower(),  # type: ignore[arg-type]
                name=m.group(2),
            )
            blocks.append(current)
            continue

        if current is not None and "  " in stripped:
            pair = _parse_field_line(stripped)
            if pair:
                current.fields[pair[0]] = pair[1]

    return blocks


def _resolve_jail_and_exec(
    process_name: str,
    known_jails: set[str],
) -> tuple[str, str]:
    best_jail: str | None = None
    best_exec: str | None = None

    for jail in known_jails:
        prefix = jail + "-"
        if process_name.startswith(prefix) and (best_jail is None or len(jail) > len(best_jail)):
            best_jail = jail
            best_exec = process_name[len(prefix) :]

    if best_jail is not None and best_exec is not None:
        return best_jail, best_exec

    if "-" in process_name:
        return process_name.rsplit("-", 1)  # type: ignore[return-value]

    return process_name, process_name


def _extract_mem(raw: str | None) -> str | None:
    if not raw:
        return None
    if "[" in raw:
        return raw.split("[")[-1].rstrip("]").strip() or None

    return raw.strip() or None


def _interpret(blocks: list[MonitBlock]) -> dict[str, MonitJailStatus]:
    results: dict[str, MonitJailStatus] = {}

    for block in blocks:
        if block.type == "system" and block.name.endswith("-system"):
            jail_name = block.name.rsplit("-system", 1)[0]
            status_val = block.fields.get("status", "unknown").strip()
            results.setdefault(jail_name, MonitJailStatus(system_ok=None, services=[]))
            results[jail_name]["system_ok"] = status_val.upper() == "OK"

    known_jails = set(results.keys())

    for block in blocks:
        if block.type != "process":
            continue

        jail_name, exec_name = _resolve_jail_and_exec(block.name, known_jails)
        entry = results.setdefault(jail_name, MonitJailStatus(system_ok=None, services=[]))
        known_jails.add(jail_name)

        status_val = block.fields.get("status", "unknown").strip().lower()
        cpu_raw = block.fields.get("cpu total") or block.fields.get("cpu")

        entry["services"].append(
            MonitService(
                name=exec_name,
                status=status_val,
                cpu=cpu_raw.strip() if cpu_raw else None,
                mem=_extract_mem(block.fields.get("memory")),
                uptime=block.fields.get("uptime", "").strip() or None,
            )
        )

    return results


def parse_monit_status(raw: str) -> dict[str, MonitJailStatus]:
    return _interpret(_tokenize(raw))



from jailrun.ansible import run_playbook
from jailrun.config import derive_plan, save_state
from jailrun.misc import lock
from jailrun.network import get_ssh_kw, wait_for_ssh
from jailrun.qemu import vm_is_running
from jailrun.schemas import ExecPlan, JailPlan, Plan, State
from jailrun.settings import Settings
from jailrun.ui import err, ok, warn


def down(state: State, settings: Settings, *, names: list[str] | None = None) -> None:
    with lock(settings.state_file):
        _down(state=state, settings=settings, names=names)


def _down(state: State, *, settings: Settings, names: list[str] | None = None) -> None:
    if not names:
        warn("No jails selected.")
        return

    alive, _ = vm_is_running(settings.pid_file)
    if not alive:
        err("VM is not running. Run 'jrun start' first.")
        raise SystemExit(1)

    new_state = state.model_copy(deep=True)

    requested = list(dict.fromkeys(names))
    removed: list[str] = []
    skipped: list[str] = []

    for name in requested:
        if name in new_state.jails:
            del new_state.jails[name]
            removed.append(name)
        else:
            skipped.append(name)

    for name in skipped:
        warn(f"Jail '{name}' not in state — skipping.")

    if not removed:
        warn("No matching jails found in state.")
        return

    plan = derive_plan(state, new_state)

    ssh_kw = get_ssh_kw(settings, state)
    wait_for_ssh(ssh_kw, silent=True)

    run_playbook("jail-forwards.yml", plan=plan, settings=settings, state=new_state)

    dns_jails = [
        JailPlan(name=name, release=j.release, ip=j.ip, base=j.base) for name, j in new_state.jails.items() if j.ip
    ]
    run_playbook(
        "jail-dns.yml",
        plan=Plan(jails=dns_jails),
        settings=settings,
        state=new_state,
    )

    stop_jails = [
        JailPlan(name=name, release=state.jails[name].release, ip=state.jails[name].ip, base=state.jails[name].base)
        for name in removed
        if name in state.jails
    ]
    stop_execs = [
        ExecPlan(name=en, jail=jn, cmd=e.cmd, dir=e.dir, env=e.env, healthcheck=e.healthcheck)
        for jn, j in state.jails.items()
        if jn in removed
        for en, e in j.execs.items()
    ]
    if stop_jails:
        run_playbook(
            "jail-stop.yml",
            plan=Plan(jails=stop_jails, execs=stop_execs),
            settings=settings,
            state=state,
        )

    run_playbook("jail-teardown.yml", plan=plan, settings=settings, state=new_state)

    run_playbook("vm-mounts.yml", plan=plan, settings=settings, state=new_state)

    save_state(state=new_state, state_file=settings.state_file)

    ok(f"Removed: {', '.join(removed)}.")



import hashlib
import re
from pathlib import Path

import httpx
from rich.progress import (
    BarColumn,
    DownloadColumn,
    Progress,
    TextColumn,
    TimeRemainingColumn,
    TransferSpeedColumn,
)

from jailrun.ui import info, ok

CHUNK_SIZE = 1024 * 1024


def sha512_file(path: Path) -> str:
    h = hashlib.sha512()
    with path.open("rb") as f:
        for chunk in iter(lambda: f.read(CHUNK_SIZE), b""):
            h.update(chunk)
    return h.hexdigest()


def fetch_checksum(bsd_image_url: str, bsd_image_checksum_url: str) -> str:
    r = httpx.get(bsd_image_checksum_url, follow_redirects=True, timeout=30)
    r.raise_for_status()
    image_xz = Path(bsd_image_url).name
    pattern = re.compile(rf"^SHA512 \({re.escape(image_xz)}\) = ([0-9a-fA-F]+)$")
    for line in r.text.splitlines():
        m = pattern.match(line.strip())
        if m:
            return m.group(1)
    raise RuntimeError("Checksum not found")


def download(bsd_image_url: str, bsd_image_checksum_url: str, *, target_dir: Path) -> None:
    image_xz_name = Path(bsd_image_url).name
    out_path = target_dir / image_xz_name
    tmp_path = out_path.with_suffix(out_path.suffix + ".tmp")

    if out_path.exists():
        info("Image already downloaded.")
        return

    info("Fetching checksum…")
    expected = fetch_checksum(bsd_image_url, bsd_image_checksum_url)

    downloaded = tmp_path.stat().st_size if tmp_path.exists() else 0
    headers = {"Range": f"bytes={downloaded}-"} if downloaded else {}

    with httpx.stream("GET", bsd_image_url, headers=headers, follow_redirects=True, timeout=None) as r:
        r.raise_for_status()
        if downloaded and r.status_code == 200:
            tmp_path.unlink(missing_ok=True)
            downloaded = 0
        total = int(r.headers.get("Content-Length", 0)) + downloaded
        mode = "ab" if downloaded else "wb"

        with Progress(
            TextColumn("[dim]{task.description}[/dim]"),
            BarColumn(),
            DownloadColumn(),
            TransferSpeedColumn(),
            TimeRemainingColumn(),
        ) as progress:
            task = progress.add_task(image_xz_name, total=total, completed=downloaded)
            with tmp_path.open(mode) as f:
                for chunk in r.iter_bytes(CHUNK_SIZE):
                    if chunk:
                        f.write(chunk)
                        progress.update(task, advance=len(chunk))

    if sha512_file(tmp_path) != expected:
        tmp_path.unlink(missing_ok=True)
        raise RuntimeError("Checksum mismatch")

    tmp_path.replace(out_path)
    ok(f"Downloaded {out_path.name}.")



from rich.padding import Padding
from rich.table import Table
from rich.text import Text
from rich.tree import Tree

from jailrun.ui import con, err

from .sparkline import SampleHistory
from .types import DEFAULT_SCOPES, JailRow, Scope, StatusInfo


def _jail_labels(j: JailRow) -> tuple[Text, Text]:
    name, st = j["name"], j["state"]
    if j.get("stale"):
        return (
            Text.assemble((name, "bold"), ("  stale", "yellow")),
            Text("missing", style="yellow"),
        )

    state_style = "green" if st.lower() == "up" else "red"

    if not j["managed"]:
        return (
            Text.assemble((name, "bold"), ("  unmanaged", "dim white")),
            Text(st.lower(), style=state_style),
        )

    return Text(name, style="bold"), Text(st.lower(), style=state_style)


def jail_header(j: JailRow) -> Text:
    name_t, state_t = _jail_labels(j)
    return Text.assemble(("  ", "")) + name_t + Text("  ") + state_t


def _multiline_cell(items: list[str], style: str = "dim white") -> Text:
    if not items:
        return Text("—", style="dim white")
    return Text("\n".join(items), style=style)


def _svc_summary_cell(j: JailRow) -> Text:
    monit = j.get("monit")
    if not monit or not monit["services"]:
        return Text("—", style="dim white")

    result = Text("")

    for i, svc in enumerate(monit["services"]):
        if i > 0:
            result.append("\n")
        svc_style = "green" if svc["status"] == "ok" else "red"
        result.append_text(
            Text.assemble(
                (svc["name"], "dim white"),
                (" ", ""),
                (svc["status"], svc_style),
            )
        )

    return result


def _add_tree_kv_rows(
    node: Tree,
    items: list[str],
    label: str,
    plural_label: str | None = None,
    style: str = "dim white",
) -> None:
    if items:
        for i, item in enumerate(items):
            prefix = f"{label:<8}" if i == 0 else " " * 8
            node.add(Text.assemble((prefix, "dim cyan"), (item, style)))
    else:
        fallback = plural_label or label + "s"
        node.add(Text.assemble((f"{fallback:<8}", "dim cyan"), ("—", "dim white")))


def _build_header(data: StatusInfo) -> Text:
    return Text.assemble(
        ("  ● ", "bold green"),
        ("VM", "bold white"),
        ("  running", "green"),
        (f"  on {data['ssh_host']}:{data['ssh_port']}", "dim white"),
        (f"  (pid {data['pid']})", "dim white"),
    )


def _build_vitals(data: StatusInfo) -> Table:
    vitals = Table.grid(padding=(0, 3, 0, 0))
    vitals.add_column(style="dim cyan", no_wrap=True, min_width=8)
    vitals.add_column()

    if data["uptime"]:
        vitals.add_row("uptime", data["uptime"].strip())
    if data["disk_free"] and data["disk_total"]:
        vitals.add_row(
            "disk",
            Text.assemble((data["disk_free"], "bold green"), (f" free of {data['disk_total']}", "dim white")),
        )
    if data["mem_total"] is not None and data["mem_usable"] is not None:
        vitals.add_row(
            "memory",
            Text.assemble(
                (f"{data['mem_usable']:.1f} GB", "bold green"),
                (f" usable / {data['mem_total']:.1f} GB total", "dim white"),
            ),
        )

    return vitals


def _add_vitals_to_tree(root: Tree, data: StatusInfo) -> None:
    if data["uptime"]:
        root.add(Text.assemble(("uptime  ", "dim cyan"), data["uptime"].strip()))
    if data["disk_free"] and data["disk_total"]:
        root.add(
            Text.assemble(
                ("disk    ", "dim cyan"),
                (data["disk_free"], "bold green"),
                (f" free of {data['disk_total']}", "dim white"),
            )
        )
    if data["mem_total"] is not None and data["mem_usable"] is not None:
        root.add(
            Text.assemble(
                ("memory  ", "dim cyan"),
                (f"{data['mem_usable']:.1f} GB", "bold green"),
                (f" usable / {data['mem_total']:.1f} GB total", "dim white"),
            )
        )


def build_jail_info(j: JailRow) -> Table:
    info = Table.grid(padding=(0, 3, 0, 0))
    info.add_column(style="dim cyan", no_wrap=True, min_width=8)
    info.add_column()

    for label, items, fallback in (
        ("ip", j["ips"], "ip"),
        ("port", j["ports"], "ports"),
        ("mount", j["mounts"], "mounts"),
    ):
        if items:
            for i, item in enumerate(items):
                info.add_row(label if i == 0 else "", Text(item, style="dim white"))
        else:
            info.add_row(fallback, Text("—", style="dim white"))

    return info


def build_jail_table(j: JailRow, *, history: SampleHistory | None = None) -> Table:
    monit = j.get("monit")
    if not monit or not monit["services"]:
        tbl = Table.grid()
        tbl.add_row(Text("no monitored services", style="dim white"))
        return tbl

    has_sparks = history is not None

    tbl = Table(
        show_header=True,
        header_style="dim",
        box=None,
        show_edge=False,
        padding=(0, 3, 0, 0),
        expand=False,
        pad_edge=False,
    )
    tbl.add_column("name", no_wrap=True)
    tbl.add_column("status", no_wrap=True)
    tbl.add_column("cpu", no_wrap=True, justify="right")

    if has_sparks:
        tbl.add_column("", no_wrap=True)

    tbl.add_column("mem", no_wrap=True, justify="right")

    if has_sparks:
        tbl.add_column("", no_wrap=True)

    tbl.add_column("uptime", no_wrap=True, style="dim white")

    for svc in monit["services"]:
        svc_style = "green" if svc["status"] == "ok" else "red"
        cells: list[str | Text] = [
            Text(svc["name"], style="bold white"),
            Text(svc["status"], style=svc_style),
            Text(svc["cpu"] or "—", style="white" if svc["cpu"] else "dim white"),
        ]
        if history is not None:
            cells.append(history.cpu_spark(j["name"], svc["name"]))

        cells.append(Text(svc["mem"] or "—", style="white" if svc["mem"] else "dim white"))

        if history is not None:
            cells.append(history.mem_spark(j["name"], svc["name"]))

        cells.append(Text(svc["uptime"] or "—", style="dim white"))
        tbl.add_row(*cells)

    return tbl


def find_jail(data: StatusInfo, jail_name: str) -> JailRow:
    for j in data["jail_rows"]:
        if j["name"] == jail_name:
            return j

    err(f"Jail '{jail_name}' not found.")

    raise SystemExit(1)


def render_overview_table(data: StatusInfo, *, scopes: frozenset[Scope] = DEFAULT_SCOPES) -> None:
    c = con()
    c.print()
    c.print(_build_header(data))
    c.print()
    c.print(Padding(_build_vitals(data), pad=(0, 0, 0, 2)))
    c.print()

    if not data["jail_rows"]:
        c.print("  [dim]no jails[/dim]")
        c.print()
        return

    show_ip = "ip" in scopes
    show_services = "services" in scopes
    has_monit = any(j.get("monit") for j in data["jail_rows"])

    tbl = Table(
        show_header=True,
        header_style="dim",
        box=None,
        show_edge=False,
        padding=(0, 3, 0, 0),
        expand=False,
        pad_edge=False,
    )
    tbl.add_column("name", no_wrap=True, vertical="top")
    tbl.add_column("state", no_wrap=True, vertical="top")

    if show_services and has_monit:
        tbl.add_column("services", no_wrap=True, vertical="top")

    if show_ip:
        tbl.add_column("ip", style="dim white", no_wrap=True, vertical="top")

    tbl.add_column("ports", style="dim white", vertical="top")
    tbl.add_column("mounts", style="dim white", vertical="top")

    for j in data["jail_rows"]:
        name_t, state_t = _jail_labels(j)
        cells: list[str | Text] = [name_t, state_t]

        if show_services and has_monit:
            cells.append(_svc_summary_cell(j))

        if show_ip:
            cells.append(_multiline_cell(j["ips"]))

        cells.append(_multiline_cell(j["ports"]))
        cells.append(_multiline_cell(j["mounts"]))

        tbl.add_row(*cells)

    c.print(Padding(tbl, pad=(0, 0, 0, 2)))
    c.print()


def render_overview_tree(data: StatusInfo, *, scopes: frozenset[Scope] = DEFAULT_SCOPES) -> None:
    c = con()
    c.print()

    show_ip = "ip" in scopes
    show_services = "services" in scopes

    root = Tree(
        Text.assemble(
            ("● ", "bold green"),
            ("VM", "bold white"),
            ("  running", "green"),
            (f"  on {data['ssh_host']}:{data['ssh_port']}", "dim white"),
            (f"  (pid {data['pid']})", "dim white"),
        )
    )
    _add_vitals_to_tree(root, data)

    jails_node = root.add(Text("jails", style="bold"))
    if not data["jail_rows"]:
        jails_node.add(Text("no jails", style="dim white"))
    else:
        for j in data["jail_rows"]:
            name_t, state_t = _jail_labels(j)
            node = jails_node.add(name_t + Text("  ") + state_t)

            if show_ip:
                _add_tree_kv_rows(node, j["ips"], "ip", plural_label="ip")

            _add_tree_kv_rows(node, j["ports"], "port", plural_label="ports")
            _add_tree_kv_rows(node, j["mounts"], "mount", plural_label="mounts")

            monit = j.get("monit")
            if show_services and monit and monit["services"]:
                for i, svc in enumerate(monit["services"]):
                    svc_style = "green" if svc["status"] == "ok" else "red"
                    node.add(
                        Text.assemble(
                            ("service " if i == 0 else "        ", "dim cyan"),
                            (svc["name"], "dim white"),
                            ("  ", ""),
                            (svc["status"], svc_style),
                        )
                    )

    c.print(Padding(root, pad=(0, 0, 0, 2)))
    c.print()


def render_jail_table(data: StatusInfo, jail_name: str) -> None:
    c = con()
    j = find_jail(data, jail_name)
    c.print()
    c.print(jail_header(j))
    c.print()
    c.print(Padding(build_jail_info(j), pad=(0, 0, 0, 2)))
    c.print()
    c.print(Padding(build_jail_table(j), pad=(0, 0, 0, 2)))
    c.print()


def render_jail_tree(data: StatusInfo, jail_name: str, *, history: SampleHistory | None = None) -> None:
    c = con()
    j = find_jail(data, jail_name)
    c.print()

    name_t, state_t = _jail_labels(j)
    root = Tree(name_t + Text("  ") + state_t)

    _add_tree_kv_rows(root, j["ips"], "ip", plural_label="ip")
    _add_tree_kv_rows(root, j["ports"], "port", plural_label="ports")
    _add_tree_kv_rows(root, j["mounts"], "mount", plural_label="mounts")

    monit = j.get("monit")
    if monit and monit["services"]:
        svc_node = root.add(Text("services", style="bold"))
        for svc in monit["services"]:
            svc_style = "green" if svc["status"] == "ok" else "red"
            label = Text.assemble(
                (svc["name"], "bold white"),
                ("  ", ""),
                (svc["status"], svc_style),
            )
            node = svc_node.add(label)

            if svc["cpu"]:
                line = Text.assemble(("cpu     ", "dim cyan"), (svc["cpu"], "white"))
                if history:
                    line.append("  ")
                    line.append_text(history.cpu_spark(j["name"], svc["name"]))
                node.add(line)

            if svc["mem"]:
                line = Text.assemble(("mem     ", "dim cyan"), (svc["mem"], "white"))
                if history:
                    line.append("  ")
                    line.append_text(history.mem_spark(j["name"], svc["name"]))
                node.add(line)

            if svc["uptime"]:
                node.add(Text.assemble(("uptime  ", "dim cyan"), (svc["uptime"], "dim white")))

    c.print(Padding(root, pad=(0, 0, 0, 2)))
    c.print()



from typing import Literal, NotRequired, TypedDict

Scope = Literal["ip", "services"]
ALL_SCOPES: tuple[Scope, ...] = ("ip", "services")
DEFAULT_SCOPES: frozenset[Scope] = frozenset()


class RawJail(TypedDict):
    private_name: str
    state: str
    ip: str


class DiskStats(TypedDict):
    disk_free: str | None
    disk_total: str | None


class MemStats(TypedDict):
    mem_total: float | None
    mem_usable: float | None


class MonitService(TypedDict):
    name: str
    status: str
    cpu: str | None
    mem: str | None
    uptime: str | None


class MonitJailStatus(TypedDict):
    system_ok: bool | None
    services: list[MonitService]


class JailRow(TypedDict):
    name: str
    state: str
    ips: list[str]
    managed: bool
    ports: list[str]
    mounts: list[str]
    stale: NotRequired[bool]
    monit: NotRequired[MonitJailStatus]


class StatusInfo(TypedDict):
    pid: int
    ssh_host: str
    ssh_port: int
    uptime: str | None
    disk_free: str | None
    disk_total: str | None
    mem_total: float | None
    mem_usable: float | None
    jail_rows: list[JailRow]



import shlex
import subprocess

import typer

from jailrun.network import get_ssh_kw, jail_ssh_cmd, ssh_cmd, wait_for_ssh
from jailrun.qemu import vm_is_running
from jailrun.schemas import State
from jailrun.settings import Settings
from jailrun.ui import err


def ssh(state: State, settings: Settings, jail_name: str | None = None) -> None:
    """Open an interactive SSH session to the VM or a jail."""
    alive, _ = vm_is_running(settings.pid_file)
    if not alive:
        err("VM is not running. Run 'jrun start' first.")
        raise typer.Exit(1)

    ssh_kw = get_ssh_kw(settings, state)
    wait_for_ssh(ssh_kw)

    if jail_name:
        if jail_name not in state.jails:
            err(f"Jail '{jail_name}' not found in state.")
            raise typer.Exit(1)

        jail_ip = state.jails[jail_name].ip
        if not jail_ip:
            err(f"Jail '{jail_name}' has no IP assigned.")
            raise typer.Exit(1)

        command = jail_ssh_cmd(args=[], jail_ip=jail_ip, tty=True, ssh_kw=ssh_kw)
    else:
        command = ssh_cmd(args=[], tty=True, ssh_kw=ssh_kw)

    subprocess.run(command, check=False)


def run_cmd(state: State, settings: Settings, jail_name: str, cmd: list[str]) -> None:
    """Execute a command inside a jail over SSH."""
    alive, _ = vm_is_running(settings.pid_file)
    if not alive:
        err("VM is not running. Run 'jrun start' first.")
        raise typer.Exit(1)

    ssh_kw = get_ssh_kw(settings, state)
    wait_for_ssh(ssh_kw)

    if jail_name not in state.jails:
        err(f"Jail '{jail_name}' not found in state.")
        raise typer.Exit(1)

    jail_ip = state.jails[jail_name].ip
    if not jail_ip:
        err(f"Jail '{jail_name}' has no IP assigned.")
        raise typer.Exit(1)

    remote_cmd = [shlex.join(cmd)]
    command = jail_ssh_cmd(args=remote_cmd, jail_ip=jail_ip, tty=True, ssh_kw=ssh_kw)

    result = subprocess.run(command, check=False)
    if result.returncode != 0:
        err(f"Command exited with code {result.returncode}.")
        raise typer.Exit(result.returncode)



from functools import cache
from pathlib import Path

from jinja2 import Environment, FileSystemLoader

from jailrun import PACKAGE_DIR


@cache
def build_jinja_env(templates_dir: Path | None = None) -> Environment:
    jinja_env = Environment(
        loader=FileSystemLoader(templates_dir or PACKAGE_DIR / "files"),
        autoescape=True,
    )
    return jinja_env



from pathlib import Path

import typer

from jailrun.ansible import run_playbook
from jailrun.cmd.stop import stop_vm
from jailrun.config import (
    derive_plan,
    derive_qemu_fwds,
    derive_qemu_shares,
    parse_config,
    resolve_base,
    save_state,
)
from jailrun.misc import lock
from jailrun.network import get_ssh_kw, resolve_ssh_port, wait_for_ssh
from jailrun.qemu import QemuMode, launch_vm, prepare_disk, vm_is_running
from jailrun.remote import fetch_remote_playbook
from jailrun.schemas import BaseState, LocalSetupStep, RemoteSetupStep, State
from jailrun.settings import Settings
from jailrun.ui import err, info, warn


def start(
    base_config: Path | None,
    state: State,
    settings: Settings,
    *,
    provision: bool = False,
    mode: QemuMode = QemuMode.SERVER,
) -> None:
    with lock(settings.state_file):
        _start_vm(
            base_config=base_config,
            state=state,
            settings=settings,
            provision=provision,
            mode=mode,
        )


def _start_vm(
    base_config: Path | None,
    state: State,
    settings: Settings,
    *,
    provision: bool = False,
    mode: QemuMode = QemuMode.SERVER,
) -> None:
    alive, pid = vm_is_running(settings.pid_file)
    if alive:
        warn(f"VM is already running (pid {pid}).")
        raise typer.Exit(0)

    vars_path = settings.disk_dir / "OVMF_VARS.fd"
    if vars_path.exists():
        vars_path.unlink()

    prepare_disk(settings)

    needs_base = not settings.state_file.exists()

    new_state = state.model_copy(deep=True)

    if base_config is None and new_state.base.is_empty():
        new_state.base = BaseState()

    if base_config and not base_config.exists():
        err(f"Base config not found: {base_config}")
        raise typer.Exit(1)

    if base_config:
        parsed = parse_config(base_config)
        if parsed.base:
            info(f"Loaded base config from {base_config.name}")
            config_base = base_config.parent.resolve()
            state.base = resolve_base(parsed.base, config_base)

    new_state.ssh_port = resolve_ssh_port(state=new_state, settings=settings)
    new_state.launched_fwds = derive_qemu_fwds(new_state)
    new_state.launched_shares = derive_qemu_shares(new_state)

    launch_vm(state=new_state, mode=QemuMode.SERVER, settings=settings)
    save_state(state=new_state, state_file=settings.state_file)

    ssh_kw = get_ssh_kw(settings, new_state)
    wait_for_ssh(ssh_kw)

    if needs_base or provision:
        run_playbook("base.yml", settings=settings, state=new_state)
        run_playbook("vm-dns-bootstrap.yml", settings=settings, state=new_state)
        if settings.mesh_network:
            run_playbook(
                "vm-yggdrasil.yml",
                extra_vars={
                    "relay_peer_uri": settings.relay_peer_uri,
                    "relay_peer_port": settings.relay_peer_port,
                    "relay_pubkey": settings.relay_pubkey,
                },
                settings=settings,
                state=new_state,
            )

    for step in new_state.base.setup.values():
        if step.type == "ansible":
            if isinstance(step, RemoteSetupStep):
                playbook_path = fetch_remote_playbook(
                    step.url,
                    cache_dir=settings.playbook_cache_dir,
                )
                run_playbook(
                    str(playbook_path),
                    extra_vars=step.vars or None,
                    settings=settings,
                    state=new_state,
                )
            if isinstance(step, LocalSetupStep):
                run_playbook(
                    step.file,
                    extra_vars=step.vars or None,
                    settings=settings,
                    state=new_state,
                )

    if new_state.base.mounts or new_state.base.forwards:
        plan = derive_plan(state, new_state)
        run_playbook("vm-mounts.yml", plan=plan, settings=settings, state=new_state)

    save_state(state=new_state, state_file=settings.state_file)

    if mode in {QemuMode.TTY, QemuMode.GRAPHIC}:
        info(f"Restarting VM in {mode} mode…")
        stop_vm(settings)
        launch_vm(state=new_state, mode=mode, settings=settings)
